/*-
 * Copyright (C)2007 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: time.cc,v 1.3 2007/03/08 23:30:10 babolo Exp $
 */

#define POSTGRES_EPOCH_JDATE 2451545 /* == date2j(2000, 1, 1) */
#define UNIX_EPOCH_JDATE     2440588 /* == date2j(1970, 1, 1) */
#define TIME_PREC_INV        1000000.0
#define DT_NOBEGIN           (-(int64_t)0x7fffffffffffffffLL - 1)
#define DT_NOEND             ((int64_t)0x7fffffffffffffffLL)
#define INVALID_ABSTIME      ((time_t)0x7FFFFFFE)
#define NOSTART_ABSTIME      ((time_t)INT_MIN)
#define NOEND_ABSTIME        ((time_t)0x7FFFFFFC)
#define JROUND(j) (rint(((float64_t)(j)) * TIME_PREC_INV) / TIME_PREC_INV)
#define TMODULO(t,q,u)                                        \
    do {                                                      \
        q = (((t) < 0) ? ceil((t) / (u)) : floor((t) / (u))); \
        if  ((q) != 0) t -= rint((q) * (u));                  \
    } while(0)

typedef union {
    int64_t  i;
    float64_t d;
} Timestamp;

typedef struct {
    Timestamp time;    /* all time units other than months and years */
    int32_t   month;   /* months and years, after time for alignment */
} Interval;

typedef struct {
    Timestamp time;    /* all time units other than months and years */
    int32_t   zone;    /* numeric time zone, in seconds              */
} TimeTzADT;

static void
j2date(int jd, int *year, int *month, int *day) {
    unsigned int julian;
    unsigned int quad;
    unsigned int extra;
    int          y;

    julian = jd;
    julian += 32044;
    quad = julian / 146097;
    extra = (julian - quad * 146097) * 4 + 3;
    julian += 60 + quad * 3 + extra / 146097;
    quad = julian / 1461;
    julian -= quad * 1461;
    y = julian * 4 / 1461;
    julian = ((y != 0) ? ((julian + 305) % 365) : ((julian + 306) % 366)) + 123;
    y += quad * 4;
    *year = y - 4800;
    quad = julian * 2141 / 65536;
    *day = julian - 7834 * quad / 256;
    *month = (quad + 10) % 12 + 1;
}   /* j2date() */

static void
interval2tm(PGsaDumpCtl *ctl, Interval span, struct tm *tm, Timestamp *fsec) {
    if  (span.month != 0) {
        tm->tm_year = span.month / 12;
        tm->tm_mon = span.month % 12;
    } else {
        tm->tm_year = 0;
        tm->tm_mon = 0;
    }
    if  (ctl->flags & PGsaDumpTINT) {
#define TIME (span.time.i)
        MODULO(TIME, tm->tm_mday, (int64_t)86400000000LL);
        MODULO(TIME, tm->tm_hour, (int64_t)3600000000LL);
        MODULO(TIME, tm->tm_min, (int64_t)60000000LL);
        MODULO(TIME, tm->tm_sec, (int64_t)1000000LL);
        (*fsec).i = TIME;
#undef TIME
    } else {
#define TIME (span.time.d)
        TMODULO(TIME, tm->tm_mday, 86400e0);
        TMODULO(TIME, tm->tm_hour, 3600e0);
        TMODULO(TIME, tm->tm_min, 60e0);
        TMODULO(TIME, tm->tm_sec, 1e0);
        (*fsec).d = TIME;
#undef TIME
}   }

static void
time2tm(PGsaDumpCtl *ctl, Timestamp time, struct tm *tm, Timestamp *fsec) {
    if  (ctl->flags & PGsaDumpTINT) {
#define TIME (time.i)
        MODULO(TIME, tm->tm_hour, (int64_t)3600000000LL);
        MODULO(TIME, tm->tm_min, (int64_t)60000000LL);
        MODULO(TIME, tm->tm_sec, (int64_t)1000000LL);
        (*fsec).i = TIME;
#undef TIME
    } else {
#define TIME (time.d)
        TMODULO(TIME, tm->tm_hour, 3600e0);
        TMODULO(TIME, tm->tm_min, 60e0);
        TMODULO(TIME, tm->tm_sec, 1e0);
        (*fsec).d = TIME;
#undef TIME
}   }

static void
dt2time(PGsaDumpCtl *ctl, Timestamp jd, int *hour, int *min, int *sec, Timestamp *fsec) {
    if  (ctl->flags & PGsaDumpTINT) {
#define TIME (jd.i)
        MODULO(TIME, *hour, (int64_t)3600000000LL);
        MODULO(TIME, *min, (int64_t)60000000LL);
        MODULO(TIME, *sec, (int64_t)1000000LL);
        (*fsec).i = TIME;
#undef TIME
    } else {
#define TIME (jd.d)
        *hour = (TIME / 3600);
        TIME -= ((*hour) * 3600);
        *min = (TIME / 60);
        TIME -= ((*min) * 60);
        *sec = TIME;
        (*fsec).d = JROUND(TIME - *sec);
#undef TIME
}   }   /* dt2time() */

static int
timestamp2tm(PGsaDumpCtl *ctl, Timestamp dt, struct tm *tm, Timestamp *fsec) {
    int       datei;
    float64_t dated;
    Timestamp time;

    if  (ctl->flags & PGsaDumpTINT) {
#define TIME (time.i)
        TIME = dt.i;
        MODULO(TIME, datei, (int64_t)86400000000LL);
        if  (TIME < (int64_t)0LL) {
            TIME += (int64_t)86400000000LL;
            datei -= 1;
        }
#undef TIME
    } else {
#define TIME (time.d)
        TIME = dt.d;
        TMODULO(TIME, dated, 86400e0);
        if  (TIME < 0) {
            TIME += 86400;
            dated -= 1;
        }
        datei = (int)dated;
#undef TIME
    }
    /* Julian day routine does not work for negative Julian days */
    if  (datei < -POSTGRES_EPOCH_JDATE) {
        errno = EINVAL;
        ifBLIN_QV2(ctl->flags) warnx("Illegal Julian date");
        return(EX_DATAERR);
    }
    /* add offset to go from J2000 back to standard Julian date */
    datei += POSTGRES_EPOCH_JDATE;
    j2date(datei, &tm->tm_year, &tm->tm_mon, &tm->tm_mday);
    dt2time(ctl, time, &tm->tm_hour, &tm->tm_min, &tm->tm_sec, fsec);
    tm->tm_isdst = -1;
    return(EX_OK);
}

static int
EncodeInterval(PGsaDumpCtl *ctl, struct tm *tm, Timestamp fsec) {
    int  is_before = 0;
    int  is_nonzero = 0;

    if  (tm->tm_year != 0) {
        printf("%d year%s", tm->tm_year, ((tm->tm_year != 1) ? "s" : ""));
        is_before = (tm->tm_year < 0);
        is_nonzero = 1;
    }
    if  (tm->tm_mon != 0) {
        printf( "%s%s%d mon%s", (is_nonzero ? " " : "")
              , ((is_before && (tm->tm_mon > 0)) ? "+" : "")
              , tm->tm_mon
              , ((tm->tm_mon != 1) ? "s" : "")
              );
        is_before = (tm->tm_mon < 0);
        is_nonzero = 1;
    }
    if  (tm->tm_mday != 0) {
        printf( "%s%s%d day%s", (is_nonzero ? " " : "")
              , ((is_before && (tm->tm_mday > 0)) ? "+" : "")
              , tm->tm_mday
              , ((tm->tm_mday != 1) ? "s" : "")
              );
        is_before = (tm->tm_mday < 0);
        is_nonzero = 1;
    }
    if  (  (!is_nonzero) || (tm->tm_hour != 0) || (tm->tm_min != 0) || (tm->tm_sec != 0)
        || ((ctl->flags & PGsaDumpTINT) ? (fsec.i != 0) : (fsec.d != 0))
        ) {
        int minus = (  (tm->tm_hour < 0) || (tm->tm_min < 0) || (tm->tm_sec < 0)
                    || ((ctl->flags & PGsaDumpTINT) ? (fsec.i < 0) : (fsec.d < 0))
                    );

        printf( "%s%s%02d:%02d", (is_nonzero ? " " : "")
              , (minus ? "-" : (is_before ? "+" : ""))
              , abs(tm->tm_hour), abs(tm->tm_min)
              );
        /* Mark as "non-zero" since the fields are now filled in */
        is_nonzero = 1;
        /* need fractional seconds? */
        if  ((ctl->flags & PGsaDumpTINT) ? (fsec.i != 0) : (fsec.d != 0)) {
            if  (ctl->flags & PGsaDumpTINT) {
                printf(":%02d.%06d", abs(tm->tm_sec), (int)((fsec.i >= 0) ? fsec.i : -(fsec.i)));
            } else {
                fsec.d += tm->tm_sec;
                printf(":%013.10f", fabs(fsec.d));
            }
        } else {
            printf(":%02d", abs(tm->tm_sec));
    }   }
    /* identically zero? then put in a unitless zero... */
    if  (!is_nonzero) {
        printf("0");
    }
    return(EX_OK);
}   /* EncodeInterval() */

static int
EncodeDateOnly(PGsaDumpCtl *ctl, struct tm *tm) {
    if  ((tm->tm_mon < 1) || (tm->tm_mon > 12)) {
        errno = EINVAL;
        ifBLIN_QV2(ctl->flags) warnx("Illegal date");
        return(EX_DATAERR);
    }
    if  (tm->tm_year > 0) {
        printf("%04d-%02d-%02d", tm->tm_year, tm->tm_mon, tm->tm_mday);
    } else {
        printf("%04d-%02d-%02d %s", -(tm->tm_year - 1), tm->tm_mon, tm->tm_mday, "BC");
    }
    return(EX_OK);
}       /* EncodeDateOnly() */

static int
EncodeTimeOnly(PGsaDumpCtl *ctl, struct tm *tm, Timestamp fsec, int *tzp) {
    if  ((tm->tm_hour < 0) || (tm->tm_hour > 24)) {
        errno = EINVAL;
        ifBLIN_QV2(ctl->flags) warnx("Illegal time");
        return(EX_DATAERR);
    }
    printf("%02d:%02d", tm->tm_hour, tm->tm_min);
    /*
     * Print fractional seconds if any.  The field widths here should be
     * at least equal to the larger of MAX_TIME_PRECISION and
     * MAX_TIMESTAMP_PRECISION.
     */
    if  (ctl->flags & PGsaDumpTINT) {
        if  (fsec.i != 0) printf(":%02d.%06lld", tm->tm_sec, fsec.i);
          else printf(":%02d", tm->tm_sec);
    } else {
        if  (fsec.d != 0) printf(":%09.6f", tm->tm_sec + fsec.d);
          else printf(":%02d", tm->tm_sec);
    }
    if  (tzp) {
        int hour, min;

        hour = -(*tzp / 3600);
        min = ((abs(*tzp) / 60) % 60);
        printf(((min != 0) ? "%+03d:%02d" : "%+03d"), hour, min);
    }
    return(EX_OK);
}   /* EncodeTimeOnly() */

static int
EncodeDateTime(PGsaDumpCtl *ctl, struct tm *tm, Timestamp fsec) {
    if  ((tm->tm_mon < 1) || (tm->tm_mon > 12) || (tm->tm_hour < 0) || (tm->tm_hour > 24)) {
        errno = EINVAL;
        ifBLIN_QV2(ctl->flags) warnx("Illegal datetime");
        return(EX_DATAERR);
    }
    printf( "%04d-%02d-%02d %02d:%02d"
          , ((tm->tm_year > 0) ? tm->tm_year : -(tm->tm_year - 1))
          , tm->tm_mon, tm->tm_mday, tm->tm_hour, tm->tm_min
          );
   /*
    * Print fractional seconds if any.  The field widths here
    * should be at least equal to MAX_TIMESTAMP_PRECISION.
    *
    * In float mode, don't print fractional seconds before 1 AD,
    * since it's unlikely there's any precision left ...
    */
    if  (ctl->flags & PGsaDumpTINT) {
        if  (fsec.i != 0) printf(":%02d.%06lld", tm->tm_sec, fsec.i);
          else printf(":%02d", tm->tm_sec);
    } else {
        if  ((fsec.d != 0) && (tm->tm_year > 0)) printf(":%09.6f", tm->tm_sec + fsec.d);
          else printf(":%02d", tm->tm_sec);
    }
    if  (tm->tm_year <= 0) printf(" BC"); else printf("+00");
    return(EX_OK);
}   /* EncodeDateTime() */
